@@ -2,6 +2,7 @@ use ec::hex_unless_dec;
 use ec::State;
 use regex::Regex;
 use std::cmp::min;
+use std::fs;
 use std::io;
 use std::io::Write;
 use std::num::NonZeroUsize;
@@ -64,6 +65,8 @@ m           Toggle whether or not characters are printed after bytes
 n           Toggle whether or not byte (n)umbers are printed before bytes
 o           Toggle using c(o)lor
 p           (p)rint current line of byte(s) (depending on 'W')
+P           Save (P)references to file (width, color, etc.)
+r           (r)ead preferences from a file.
 R           Toggle (R)ead-only mode
 s           Print (s)tate of toggles, 'W'idth, etc.
 S           (S)ave state to a file except the bytes you're editing.
@@ -136,7 +139,7 @@ impl Command {
         let re_search_again = Regex::new(r"^ *(?P<direction>[/?]) *$").unwrap();
         let re_search_kill = Regex::new(r"^ */(?P<bytes>[0-9a-fA-F]+)/k *$").unwrap();
         let re_search_insert = Regex::new(r"^ */(?P<bytes>[0-9a-fA-F]+)/i *$").unwrap();
-        let re_single_char_command = Regex::new(r"^ *(?P<command>[hijkmnopqRsSlLuwx]).*$").unwrap();
+        let re_single_char_command = Regex::new(r"^ *(?P<command>[hijkmnopqRrsSlLPuwx]).*$").unwrap();
         let re_range = Regex::new(r"^ *(?P<begin>[0-9a-fA-F.$]+) *, *(?P<end>[0-9a-fA-F.$]+) *(?P<the_rest>.*) *$").unwrap();
         let re_specified_index = Regex::new(r"^ *(?P<index>[0-9A-Fa-f.$]+) *(?P<the_rest>.*) *$").unwrap();
         let re_offset_index = Regex::new(r"^ *(?P<sign>[-+])(?P<offset>[0-9A-Fa-f]+) *(?P<the_rest>.*) *$").unwrap();
@@ -380,7 +383,7 @@ impl Command {
         else if is_before_context {
             let caps = caps.unwrap();
             let given = caps.name("before_context").unwrap().as_str();
-            if let Ok(before_context) = usize::from_str_radix(given, state.radix) {
+            if let Ok(before_context) = usize::from_str_radix(given, state.prefs.radix) {
               Ok(Command{
                   range: (usize::from(before_context), usize::from(before_context)),
                   command: 'T',
@@ -395,7 +398,7 @@ impl Command {
         else if is_after_context {
             let caps = caps.unwrap();
             let given = caps.name("after_context").unwrap().as_str();
-            if let Ok(after_context) = usize::from_str_radix(given, state.radix) {
+            if let Ok(after_context) = usize::from_str_radix(given, state.prefs.radix) {
               Ok(Command{
                   range: (usize::from(after_context), usize::from(after_context)),
                   command: 't',
@@ -410,7 +413,7 @@ impl Command {
         else if is_width {
             // println!("is_width");
             let caps = caps.unwrap();
-            if let Some(width) = NonZeroUsize::new(usize::from_str_radix(caps.name("width").unwrap().as_str(), state.radix).unwrap()) {
+            if let Some(width) = NonZeroUsize::new(usize::from_str_radix(caps.name("width").unwrap().as_str(), state.prefs.radix).unwrap()) {
               Ok(Command{
                   range: (usize::from(width), usize::from(width)),
                   command: 'W',
@@ -437,7 +440,7 @@ impl Command {
             // println!("is_range");
             let caps = caps.unwrap();
             let begin = number_dot_dollar(state.index, _max_index,
-                    caps.name("begin").unwrap().as_str(), state.radix);
+                    caps.name("begin").unwrap().as_str(), state.prefs.radix);
             if begin.is_err() {
                 // Why on Earth doesn't this work?
                 // return Err(begin.unwrap());
@@ -445,7 +448,7 @@ impl Command {
             }
             let begin = begin.unwrap();
             let end = number_dot_dollar(state.index, _max_index,
-                    caps.name("end").unwrap().as_str(), state.radix);
+                    caps.name("end").unwrap().as_str(), state.prefs.radix);
             if end.is_err() {
                 // Why on Earth doesn't this work?
                 // return end;
@@ -481,7 +484,7 @@ impl Command {
             // println!("is_specified_index");
             let caps = caps.unwrap();
             let specific_index = number_dot_dollar(state.index, _max_index,
-                    caps.name("index").unwrap().as_str(), state.radix);
+                    caps.name("index").unwrap().as_str(), state.prefs.radix);
             if specific_index.is_err() {
                 // Why on Earth doesn't this work?
                 // return specific_index;
@@ -504,7 +507,7 @@ impl Command {
                     range: (specific_index, specific_index),
                     command:
                         if command == 'p' {
-                            'P'
+                            '☃'
                         }
                         else {
                           command
@@ -518,7 +521,7 @@ impl Command {
             // println!("is_specified_index");
             let caps = caps.unwrap();
             let as_string = caps.name("offset").unwrap().as_str();
-            let index_offset = usize::from_str_radix(as_string, state.radix);
+            let index_offset = usize::from_str_radix(as_string, state.prefs.radix);
             if index_offset.is_err() {
                 return Err(format!("{} is not a number", as_string));
             }
@@ -663,6 +666,24 @@ pub fn update_filename(state: &mut ec::State) {
 }
 
 
+pub fn load_state_from_file(state: &mut ec::State) {
+    let filename = read_string_from_user(Some(
+                    "Enter filename from which to load state: "));
+    if filename.is_ok() {
+        let new_state = State::from_filename(&filename.unwrap());
+        if new_state.is_ok() {
+            *state = new_state.unwrap();
+        }
+        else {
+            println!("? {:?}", new_state);
+        }
+    }
+    else {
+        println!("? {:?}", filename);
+    }
+}
+
+
 pub fn load_new_file(state: &mut ec::State) {
 	if state.unsaved_changes {
 		let unsaved_prompt = "You have unsaved changes.  Carry on? (y/n): ";
@@ -722,6 +743,76 @@ pub fn load_new_file(state: &mut ec::State) {
 }
 
 
+pub fn load_prefs(state: &mut ec::State) {
+    let pref_path = ec::preferences_file_path();
+    let filename = read_string_from_user(Some(&format!(
+            "Enter filename from which to load preferences [{}]: ",
+                    pref_path.display())));
+    if filename.is_ok() {
+        let mut filename = filename.unwrap();
+        if filename == "" {
+            if let Some(pref_path_s) = pref_path.to_str() {
+                filename = pref_path_s.to_owned();
+            }
+            else {
+                println!("? Default path ({}) is not valid unicode.",
+                        pref_path.display());
+                return;
+            }
+        }
+
+        let result = ec::Preferences::read_from_disk(&filename);
+        if result.is_ok() {
+            state.prefs = result.unwrap();
+        }
+        else {
+            println!("? {:?}", result);
+        }
+    }
+    else {
+        println!("? {:?}", filename);
+    }
+}
+
+
+pub fn save_prefs(state: &ec::State) {
+    let pref_path = ec::preferences_file_path();
+    let filename = read_string_from_user(Some(&format!(
+            "Enter filename to save preferences [{}]: ",
+                    pref_path.display())));
+    if filename.is_ok() {
+        let mut filename = filename.unwrap();
+        if filename == "" {
+            if let Some(pref_path_s) = pref_path.to_str() {
+                filename = pref_path_s.to_owned();
+
+                /* In the default case, we're brave enough to create
+                 * the parent directory for the file if it's not root */
+                if let Some(parent_dir) = pref_path.parent() {
+                    if let Err(error) = fs::create_dir_all(parent_dir) {
+                        println!("? Couldn't create directory {} ({:?})",
+                                parent_dir.display(), error);
+                    }
+                }
+            }
+            else {
+                println!("? Default path ({}) is not valid unicode.",
+                        pref_path.display());
+                return;
+            }
+        }
+
+        let result = state.prefs.write_to_disk(&filename);
+        if let Err(error) = result {
+            println!("? {:?}", error);
+        }
+    }
+    else {
+        println!("? {:?}", filename);
+    }
+}
+
+
 pub fn save_state(state: &ec::State) {
     let filename = read_string_from_user(Some("Enter filename to save state: "));
     if filename.is_ok() {
@@ -777,19 +868,23 @@ pub fn actual_runtime(filename:&str, quiet:bool, color:bool, readonly:bool)
         -> i32 {
 
     // TODO Below here should be a function called main_loop()
-    let mut state = ec::State{
-        radix: 16,
-        filename: filename.to_owned(),
-        show_byte_numbers: true,
-        show_prompt: !quiet,
-        color: color,
-        show_chars: true,
+    let mut state = ec::State {
+        prefs: ec::Preferences {
+            radix: 16,
+            show_byte_numbers: true,
+            show_prompt: !quiet,
+            color: color,
+            show_chars: true,
+            before_context: 0,
+            after_context: 0,
+            width: NonZeroUsize::new(16).unwrap(),
+            // TODO calculate based on longest possible index
+            n_padding: "      ".to_owned(),
+        },
         unsaved_changes: (filename == ""),
+        filename: filename.to_owned(),
         readonly: readonly,
         index: 0,
-        before_context: 0,
-        after_context: 0,
-        width: NonZeroUsize::new(16).unwrap(),
         all_bytes: if filename == "" {
                 Vec::new()
             }
@@ -815,8 +910,6 @@ pub fn actual_runtime(filename:&str, quiet:bool, color:bool, readonly:bool)
                 }
             }
         ,
-        // TODO calculate based on longest possible index
-        n_padding: "      ".to_owned(),
         last_search: None,
     };
 
@@ -828,7 +921,7 @@ pub fn actual_runtime(filename:&str, quiet:bool, color:bool, readonly:bool)
     }
 
     loop {
-        if state.show_prompt {
+        if state.prefs.show_prompt {
             print!("*");
         }
         io::stdout().flush().unwrap();
@@ -928,7 +1021,7 @@ pub fn actual_runtime(filename:&str, quiet:bool, color:bool, readonly:bool)
                             continue;
                         };
 
-                        let width = usize::from(state.width);
+                        let width = usize::from(state.prefs.width);
                         let first_byte_to_show_index =
                                 state.index.saturating_sub(width);
                         state.index = first_byte_to_show_index;
@@ -960,51 +1053,37 @@ pub fn actual_runtime(filename:&str, quiet:bool, color:bool, readonly:bool)
 
                     /* Load state from a file */
                     'L' => {
-                        let filename =
-                                read_string_from_user(Some(
-                                        "Enter filename from which to load state: "));
-                        if filename.is_ok() {
-                            let new_state = State::from_filename(&filename.unwrap());
-                            if new_state.is_ok() {
-                                state = new_state.unwrap();
-                            }
-                            else {
-                                println!("? {:?}", new_state);
-                            }
-                        }
-                        else {
-                            println!("? {:?}", filename);
-                        }
+                        load_state_from_file(&mut state);
                     },
 
                     /* Toggle showing char representations of bytes */
                     'm' => {
-                        state.show_chars = !state.show_chars;
+                        state.prefs.show_chars = !state.prefs.show_chars;
                         if !quiet {
-                            println!("{}", state.show_chars);
+                            println!("{}", state.prefs.show_chars);
                         }
                     },
 
                     /* Toggle showing byte number */
                     'n' => {
-                        state.show_byte_numbers = !state.show_byte_numbers;
+                        state.prefs.show_byte_numbers = !state.prefs.show_byte_numbers;
                         if !quiet {
-                            println!("{}", state.show_byte_numbers);
+                            println!("{}", state.prefs.show_byte_numbers);
                         }
                     },
 
                     /* Toggle color */
                     'o' => {
-                        state.color = !state.color;
+                        state.prefs.color = !state.prefs.color;
                         if !quiet {
-                            println!("{}", state.color);
+                            println!("{}", state.prefs.color);
                         }
                     },
 
 
                     /* Toggle hex/dec */
                     'x' => {
-                        state.radix = if state.radix == 16 {
+                        state.prefs.radix = if state.prefs.radix == 16 {
                             10
                         }
                         else {
@@ -1021,7 +1100,7 @@ pub fn actual_runtime(filename:&str, quiet:bool, color:bool, readonly:bool)
 
                         match state.max_index() {
                             Ok(max) => {
-                                let width = usize::from(state.width);
+                                let width = usize::from(state.prefs.width);
                                 let first_byte_to_show_index = state.index + width;
                                 let last_byte_to_show_index = min(
                                         first_byte_to_show_index + width - 1,
@@ -1029,7 +1108,7 @@ pub fn actual_runtime(filename:&str, quiet:bool, color:bool, readonly:bool)
                                 if first_byte_to_show_index > max {
                                     println!("? (already showing last byte at index {})",
                                             hex_unless_dec(last_byte_to_show_index,
-                                                    state.radix));
+                                                    state.prefs.radix));
                                 }
                                 else {
                                     state.index = first_byte_to_show_index;
@@ -1043,7 +1122,7 @@ pub fn actual_runtime(filename:&str, quiet:bool, color:bool, readonly:bool)
                     }
 
                     /* Print byte(s) at one place, width long */
-                    'P' => {
+                    '☃' => {
                         if state.empty() {
                             println!("? (Empty file)");
                             continue;
@@ -1072,6 +1151,17 @@ pub fn actual_runtime(filename:&str, quiet:bool, color:bool, readonly:bool)
                         }
                     },
 
+                    /* Save preferences to a file */
+                    'P' => {
+                        save_prefs(&state);
+                    },
+
+
+                    /* Load preferences from a file */
+                    'r' => {
+                        load_prefs(&mut state);
+                    },
+
                     /* Print byte(s) at *current* place, width long */
                     'Q' => {
                         if state.empty() {
@@ -1104,12 +1194,12 @@ pub fn actual_runtime(filename:&str, quiet:bool, color:bool, readonly:bool)
 
                     /* Change after_context */
                     't' => {
-                        state.after_context = usize::from(command.range.0);
+                        state.prefs.after_context = usize::from(command.range.0);
                     },
 
                     /* Change before_context */
                     'T' => {
-                        state.before_context = usize::from(command.range.0);
+                        state.prefs.before_context = usize::from(command.range.0);
                     },
 
                     /* (u)pdate iflename */
@@ -1125,7 +1215,7 @@ pub fn actual_runtime(filename:&str, quiet:bool, color:bool, readonly:bool)
                     /* Change width */
                     'W' => {
                         if let Some(width) = NonZeroUsize::new(command.range.0) {
-                            state.width = width;
+                            state.prefs.width = width;
                         }
                     },
 
